#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "diskio.h"
#include "partition.h"
#include "ps2part.h"

static void set_date(PARTBUF *pbuf, struct ps2_part_date *p)
{
    time_t tm;
    struct tm *tp;

    time(&tm);
    tm += PS2_PART_DATE_TZONE;
    tp = gmtime(&tm);

    p->sec = tp->tm_sec;
    p->min = tp->tm_min;
    p->hour = tp->tm_hour;
    p->day = tp->tm_mday;
    p->month = tp->tm_mon + 1;
    p->year = tp->tm_year + 1900;

    pbuf->modified = 1;
}

static __u32 calc_sum(struct ps2_partition *pt)
{
    int i;
    __u32 sum = 0;
    __u32 *p = (__u32 *)pt;

    for (i = 1; i < sizeof(struct ps2_partition) / sizeof(__u32); i++)
	sum += p[i];
    return sum;
}

static sector_t get_max_sector(DISKDEV *dev)
{
    sector_t size;
    int i;

    for (i = 31; i > 0; i--) {
	size = 1 << i;
	if (size & (dev->nsect >> PS2_PART_MAXSUB_SHIFT))
	    break;
    }
    
    if (size != (dev->nsect >> PS2_PART_MAXSUB_SHIFT))
	size <<= 1;
    if (size < PS2_PART_MINSIZE)
	size = PS2_PART_MINSIZE;
    
    return size;
}

static int validate_partbuf(PARTBUF *pbuf)
{
    if (pbuf->sect == 0) {
	/* mbr check */
	if (strncmp(pbuf->part->mbr.magic, PS2_MBR_MAGIC, 32) != 0)
	    return EP_BADMBR;
	if (pbuf->part->mbr.version > PS2_PART_VERSION)
	    return EP_BADMBR;
    }

    if (pbuf->part->magic != PS2_PART_MAGIC)
	return EP_BADPART;
    if (pbuf->part->sum != calc_sum(pbuf->part))
	return EP_BADPART;

    return EP_OK;
}

static void free_partbuf(PARTBUF *pbuf)
{
    free(pbuf->part);
    free(pbuf);
}

void free_partition(PARTBUF *pbuf)
{
    PARTBUF *next;

    while (pbuf != NULL) {
	next = pbuf->next;
	free_partbuf(pbuf);
	pbuf = next;
    }
}

static PARTBUF *alloc_partbuf(DISKDEV *dev)
{
    PARTBUF *pbuf;
    int nsect = get_nsect(dev, sizeof(struct ps2_partition));

    if ((pbuf = malloc(sizeof(PARTBUF))) == NULL)
	return NULL;
    memset(pbuf, 0, sizeof(PARTBUF));
    pbuf->nsect = nsect;
    pbuf->modified = 0;

    if ((pbuf->part = malloc(nsect * dev->sector_size)) == NULL) {
	free(pbuf);
	return NULL;
    }
    memset(pbuf->part, 0, nsect * dev->sector_size);
    return pbuf;
}

static PARTBUF *make_empty_partbuf(DISKDEV *dev, PARTBUF *pbuf)
{
    if (pbuf == NULL) {
	if ((pbuf = alloc_partbuf(dev)) == NULL)
	    return NULL;
    }

    pbuf->part->sum = 0;
    pbuf->part->magic = PS2_PART_MAGIC;
    memset(pbuf->part->id, 0, PS2_PART_NID);
    memset(pbuf->part->rpwd, 0, PS2_PART_NPASSWD);
    memset(pbuf->part->fpwd, 0, PS2_PART_NPASSWD);
    pbuf->part->fstype = 0;
    pbuf->part->flag = 0;
    pbuf->part->nsub = 0;
    pbuf->part->main = 0;
    pbuf->part->number = 0;
    pbuf->part->modver = 0;
    memset(pbuf->part->name, 0, PS2_PART_NNAME);
    memset(&pbuf->part->subs, 0, sizeof(struct ps2_part_subpart) * PS2_PART_MAXSUB);

    set_date(pbuf, &pbuf->part->date);
    strncpy(pbuf->part->id, PS2_ID_EMPTY, PS2_PART_NID);
    return pbuf;
}

static PARTBUF *split_empty_partbuf(DISKDEV *dev, PARTBUF *phead,
				    PARTBUF *pbuf, sector_t size)
{
    PARTBUF *npbuf;
    sector_t nstart;

    while (pbuf->part->nsector != size) {
	if ((npbuf = make_empty_partbuf(dev, NULL)) == NULL)
	    return NULL;
	pbuf->part->nsector >>= 1;
	npbuf->part->nsector = pbuf->part->nsector;
	pbuf->modified = 1;

	npbuf->next = pbuf->next;
	npbuf->prev = pbuf;
	pbuf->next = npbuf;

	nstart = pbuf->part->start + pbuf->part->nsector;
	npbuf->sect = nstart;
	npbuf->part->start = nstart;

	npbuf->part->next = pbuf->part->next;
	npbuf->part->prev = pbuf->part->start;

	pbuf->part->next = npbuf->part->start;
	if (npbuf->next == NULL) {	/* last partition */
	    phead->part->prev = npbuf->part->start;
	    phead->modified = 1;
	} else {
	    npbuf->next->part->prev = npbuf->part->start;
	    npbuf->next->prev = npbuf;
	    npbuf->next->modified = 1;
	}
    }
    make_empty_partbuf(dev, pbuf);
    return pbuf;
}

static PARTBUF *create_aligned_partbuf(DISKDEV *dev, PARTBUF *phead,
				       PARTBUF *pbuf, sector_t size)
{
    PARTBUF *npbuf;
    sector_t nstart, nsize;

    do {
	nstart = pbuf->part->start + pbuf->part->nsector;
	for (nsize = 1; nsize != 0; nsize <<= 1)
	    if (nstart & nsize)
		break;
	if (nsize == 0 || nsize > size)
	    nsize = size;
	if ((npbuf = make_empty_partbuf(dev, NULL)) == NULL)
	    return NULL;
	npbuf->part->nsector = nsize;

	npbuf->next = pbuf->next;	/* = NULL */
	npbuf->prev = pbuf;
	pbuf->next = npbuf;
	pbuf->modified = 1;

	npbuf->sect = nstart;
	npbuf->part->start = nstart;

	npbuf->part->next = pbuf->part->next;
	npbuf->part->prev = pbuf->part->start;

	pbuf->part->next = npbuf->part->start;
	phead->part->prev = npbuf->part->start;
	phead->modified = 1;

	pbuf = npbuf;
    } while (pbuf->part->nsector < size);
    return pbuf;
}

static int alloc_partition_segment(DISKDEV *dev, PARTBUF *phead,
				   PARTBUF **ppbuf, sector_t size)
{
    PARTBUF *pbuf, *pnext, *pbfree;
    sector_t nstart;

    if ((size & (size - 1)) != 0)
	return EP_ILLEGAL;
    if (size < PS2_PART_MINSIZE ||
	size > get_max_sector(dev))
	return EP_ILLEGAL;

    pbuf = pbfree = NULL;
    for (pnext = phead; pnext != NULL; pnext = pbuf->next) {
	pbuf = pnext;
	if (pbuf->part->fstype == 0) {
	    if (pbuf->part->nsector == size) {
		make_empty_partbuf(dev, pbuf);
		*ppbuf = pbuf;
		return EP_OK;
	    } else if (pbuf->part->nsector > size) {
		if (pbfree == NULL ||
		    pbfree->part->nsector > pbuf->part->nsector)
		    pbfree = pbuf;
	    }
	}
    }
    if (pbfree != NULL) {
	if ((*ppbuf = split_empty_partbuf(dev, phead, pbfree, size)) == NULL)
	    return EP_NOMEM;
	return EP_OK;
    }

    nstart = (pbuf->part->start + pbuf->part->nsector + size - 1) & ~(size - 1);
    if (nstart + size > dev->nsect)
	return EP_FULL;

    if ((*ppbuf = create_aligned_partbuf(dev, phead, pbuf, size)) == NULL)
	return EP_NOMEM;
    return EP_OK;
}

static int free_partition_segment(DISKDEV *dev, PARTBUF *phead, PARTBUF *pbuf)
{
    PARTBUF *pprev, *pnext;

    if (pbuf->part->nsub != 0)
	return EP_BADPART;

    make_empty_partbuf(dev, pbuf);
    if (pbuf->part->next == 0) {
	/* remove the last partition */
	do {
	    pprev = pbuf->prev;
	    free_partbuf(pbuf);
	    pbuf = pprev;

	    if (pbuf == NULL)
		pbuf = phead;

	    pbuf->next = NULL;
	    pbuf->part->next = 0;
	    pbuf->modified = 1;
	    phead->part->prev = pbuf->part->start;
	    phead->modified = 1;
	} while (pbuf->part->fstype == 0);
    } else {
	/* concatenate with previous empty partitions */
	while ((pprev = pbuf->prev) != NULL) {
	    if (pprev->part->fstype != 0 ||
		pprev->part->nsector != pbuf->part->nsector ||
		(pprev->part->start & (pbuf->part->nsector * 2 - 1)) != 0)
		break;
	    pprev->part->nsector <<= 1;
	    pprev->part->next = pbuf->part->next;
	    pprev->next = pbuf->next;
	    pprev->modified = 1;

	    pbuf->next->part->prev = pprev->part->start;
	    pbuf->next->prev = pprev;

	    free_partbuf(pbuf);
	    pbuf = pprev;
	}
	/* concatenate with next empty partitions */
	while ((pnext = pbuf->next) != NULL) {
	    if (pnext->part->fstype != 0 ||
		pnext->part->nsector != pbuf->part->nsector ||
		(pbuf->part->start & (pbuf->part->nsector * 2 - 1)) != 0)
		break;
	    pbuf->part->nsector <<= 1;
	    pbuf->part->next = pnext->part->next;
	    pbuf->next = pnext->next;
	    pbuf->modified = 1;

	    if (pbuf->next == NULL) {	/* last partition */
		phead->part->prev = pbuf->part->start;
		phead->modified = 1;
	    } else {
		pbuf->next->part->prev = pbuf->part->start;
		pbuf->next->prev = pbuf;
		pbuf->next->modified = 1;
	    }

	    free_partbuf(pnext);
	}
    }
    return EP_OK;
}

PARTBUF *find_partition(PARTBUF *phead, char *id)
{
    PARTBUF *pbuf;

    for (pbuf = phead; pbuf != NULL; pbuf = pbuf->next) {
	if (!(pbuf->part->flag & PS2_PART_FLAG_SUB) &&
	    strncmp(pbuf->part->id, id, PS2_PART_NID) == 0)
	    return pbuf;
    }

    return NULL;
}

int next_avail_partition(PARTBUF *phead)
{
   int nextpart;
   char name[17];
   
   for(nextpart=1;nextpart <= LINUX_PART_MAX;nextpart++)
   {
      sprintf(name, LINUX_ID, nextpart);
      if (find_partition(phead, name) == NULL)
         break;
   }
   return nextpart;
}

int create_partition(DISKDEV *dev, PARTBUF *phead, char *id, sector_t size,
		     int fstype, char *name)
{
    PARTBUF *pbuf, *pmain;
    sector_t ssize;
    int i;

    if (find_partition(phead, id) != NULL)
	return EP_EXIST;
    if (fstype <= 1)
	return EP_ILLEGAL;
    if (size <= 0)
	return EP_ILLEGAL;

    size = (size + PS2_PART_MINSIZE - 1) & ~(PS2_PART_MINSIZE - 1);
    pmain = NULL;

    for (i = 31; i > 0; i--) {
	ssize = 1 << i;
	if (ssize & size)
	    break;
    }

    while (size > 0) {
	while (size < ssize)
	    ssize >>= 1;
	if (ssize < PS2_PART_MINSIZE) {
	    if (pmain != NULL)
		remove_partition(dev, phead, id);
	    return EP_FULL;
	}
	switch (alloc_partition_segment(dev, phead, &pbuf, ssize)) {
	case EP_OK:
	    size -= ssize;
	    pbuf->part->fstype = fstype;
	    pbuf->part->modver = MODVER;
	    memset(pbuf->part->id, 0, PS2_PART_NID);
	    if (pmain == NULL) {
		pmain = pbuf;
		pmain->part->nsub = 0;
		strncpy(pbuf->part->id, id, PS2_PART_NID);
		strncpy(pmain->part->name, name, PS2_PART_NNAME);
	    } else {
		pbuf->part->flag = PS2_PART_FLAG_SUB;
		pmain->part->subs[pmain->part->nsub].start = pbuf->part->start;
		pmain->part->subs[pmain->part->nsub].nsector = pbuf->part->nsector;
		pmain->part->nsub++;
		pbuf->part->main = pmain->part->start;
		pbuf->part->number = pmain->part->nsub;
		if (pmain->part->nsub >= PS2_PART_MAXSUB)
		    ssize = 0;
	    }
	    break;
	case EP_FULL:
	case EP_ILLEGAL:
	    ssize >>= 1;
	    break;
	case EP_NOMEM:
	default:
	    remove_partition(dev, phead, id);
	    return EP_NOMEM;
	}
    }
    return EP_OK;
}

static PARTBUF *find_subpart_sector(PARTBUF *phead, sector_t sector)
{
    PARTBUF *pbuf;

    for (pbuf = phead; pbuf != NULL; pbuf = pbuf->next) {
	if (pbuf->part->flag & PS2_PART_FLAG_SUB &&
	    pbuf->part->start == sector)
	    return pbuf;
    }

    return NULL;
}

int remove_partition(DISKDEV *dev, PARTBUF *phead, char *id)
{
    PARTBUF *pbuf, *psub;
    sector_t substart;
    int i;

    if ((pbuf = find_partition(phead, id)) == NULL)
	return EP_NOPART;
    if (pbuf->part->fstype <= 1)
	return EP_ILLEGAL;		/* cannot delete system partition */

    pbuf->modified = 1;
    for (i = pbuf->part->nsub; i > 0; i--) {
	substart = pbuf->part->subs[i - 1].start;
	if ((psub = find_subpart_sector(phead, substart)) != NULL) {
	    free_partition_segment(dev, phead, psub);
	    pbuf->part->subs[i - 1].start = 0;
	    pbuf->part->subs[i - 1].nsector = 0;
	    pbuf->part->nsub--;
	}
    }
    return free_partition_segment(dev, phead, pbuf);
}

sector_t getfree_partition(DISKDEV *dev, PARTBUF *phead)
{
    PARTBUF *pbuf;
    sector_t used, size;

    used = 0;
    size = dev->nsect & ~(PS2_PART_MINSIZE - 1);
    for (pbuf = phead; pbuf != NULL; pbuf = pbuf->next) {
	if (pbuf->part->fstype != 0)
	    used += pbuf->part->nsector;
    }

    return size - used;
}

sector_t getsize_partition(DISKDEV *dev, PARTBUF *pbuf)
{
    sector_t size;
    int i;

    size = pbuf->part->nsector;
    for (i = 0; i < pbuf->part->nsub; i++)
	size += pbuf->part->subs[i].nsector;

    return size;
}

int format_partition(DISKDEV *dev, PARTBUF **pphead)
{
    PARTBUF *pbuf;

    if ((pbuf = alloc_partbuf(dev)) == NULL)
	return EP_NOMEM;

    pbuf->sect = 0;

    pbuf->part->magic = PS2_PART_MAGIC;
    pbuf->part->nsector = PS2_PART_MINSIZE;
    pbuf->part->fstype = 1;
    set_date(pbuf, &pbuf->part->date);
    strncpy(pbuf->part->id, PS2_ID_MBR, PS2_PART_NID);
    strncpy(pbuf->part->name, PS2_ID_MBR, PS2_PART_NNAME);
    strncpy(pbuf->part->mbr.magic, PS2_MBR_MAGIC, 32);
    pbuf->part->mbr.version = PS2_PART_VERSION;
    pbuf->part->mbr.nsector = 0;
    set_date(pbuf, &pbuf->part->mbr.date);
    pbuf->part->mbr.osd_start = 0;
    pbuf->part->mbr.osd_count = 0;

    *pphead = pbuf;
    return EP_OK;
}

static int read_partbuf(DISKDEV *dev, PARTBUF **ppbuf, sector_t sect)
{
    PARTBUF *pbuf;

    if ((pbuf = alloc_partbuf(dev)) == NULL)
	return EP_NOMEM;

    pbuf->sect = sect;
    
    if (read_sector(dev, pbuf->sect, (unsigned char *)pbuf->part, pbuf->nsect) < 0) {
	free_partbuf(pbuf);
	return EP_READ;
    }

    *ppbuf = pbuf;
    return EP_OK;
}

int read_partition(DISKDEV *dev, PARTBUF **pphead)
{
    PARTBUF *phead, *ptail, *pbuf;
    sector_t sect = 0;
    int result;

    if ((result = read_partbuf(dev, &phead, sect)) < 0)
	return result;
    if ((result = validate_partbuf(phead)) < 0) {
	free_partbuf(phead);
	return result;
    }

    ptail = phead;
    sect = phead->part->next;

    while (sect != 0) {
	if ((result = read_partbuf(dev, &pbuf, sect)) < 0) {
	    free_partition(phead);
	    return result;
	}
	if ((result = validate_partbuf(pbuf)) < 0) {
	    free_partbuf(pbuf);
	    free_partition(phead);
	    return result;
	}

	ptail->next = pbuf;
	pbuf->prev = ptail;
	ptail = pbuf;
	sect = pbuf->part->next;
    }

    *pphead = phead;
    return EP_OK;
}

int write_partition(DISKDEV *dev, PARTBUF *phead)
{
    PARTBUF *pbuf = phead;
    int result;

    while (pbuf != NULL) {
	if (pbuf->modified) {
	    pbuf->part->sum = calc_sum(pbuf->part);

	    result = write_sector(dev, pbuf->sect, (unsigned char *)pbuf->part, pbuf->nsect);
	    if (result < 0)
		return EP_WRITE;
	}
	pbuf = pbuf->next;
    }
    return EP_OK;
}
