#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <linux/fs.h>

#include "diskio.h"
#include "partition.h"
#include "ps2part.h"

#define DEFAULT_DEVICE		"/dev/hda"

#define sect2mb(dev, sect)	((sect) >> (20 - (dev)->sector_shift))
#define mb2sect(dev, sect)	((sect) << (20 - (dev)->sector_shift))

#define LINUX_TYPE		0x83
#define LINUX_SWAP_TYPE		0x82
#define LINUX_REISERFS_TYPE	0x88
#define LINUX_LVM_TYPE		0x8e

#define LINE_LENGTH	256
char line_buffer[LINE_LENGTH];

char *errmsg[] = {
    "Disk read error.",
    "Disk write error.",
    "Internal error --- no memory.",
    "Bad MBR format.",
    "Bad partition format.",
    "No such partiton.",
    "Illegal parameter.",
    "Disk full.",
    "Partition ID already exists.",
};

void print_error(int error)
{
    puts(errmsg[-error - 1]);
}

void perror_fdisk(char *msg)
{
    if (msg == NULL) {
	perror("ps2fdisk");
    } else {
	fprintf(stderr, "ps2fdisk: ");
	perror(msg);
    }
}

void usage(void)
{
   printf(
      "Usage: ps2fdisk [option] [DISK]\n"
      "       Creates, partitions, and repartitions the Sony\n"
      "       PlayStation 2 Hard Disk Drive unit. When executed\n"
      "       without arguments the menued interface is launched.\n"
      "\n"
      "       options:\n"
      "         -l [DISK]            List partition table of DISK\n"
      "         -s <PARTITION>       Give PARTITION size(s) in blocks\n"
      "         -n <DISK> <SIZE>     Add new TYPE SIZE MB partition on DISK\n"
      "            <TYPE> <COMMENT>  with COMMENT. The first free partition\n"
      "                              number is used.\n"
      "         -o <DISK>            Create an empty partition table on DISK.\n"
      "         -c <DISK>            Check for partition table on DISK.\n"
      );
   exit(1);
}

void menu(void)
{
    printf(
	"Command action\n"
	"   d   delete a partition\n"
	"   n   add a new partition\n"
	"   o   create a new empty PlayStation 2 partition table\n"
	"   p   print the partition table\n"
	"   q   quit without saving changes\n"
	"   w   write table to disk and exit\n"
	);
}

void menu_expert(void)
{
    printf(
	"Command action\n"
	"   p   print the partition table\n"
	"   q   quit without saving changes\n"
	"   r   return to main menu\n"
	"   w   write table to disk and exit\n"
	);
}

char *read_line(void)
{
    char *p;

    if (fgets(line_buffer, LINE_LENGTH, stdin) == NULL)
	return NULL;
    for (p = line_buffer; *p != '\0'; p++)
	if (*p == '\r' || *p == '\n')
	    *p = '\0';
    return line_buffer;
}

int read_num(char *mesg, int min, int max)
{
    int val;
    char *p;

    printf("%s (%d-%d): ", mesg, min, max);
    if ((p = read_line()) == NULL)
	return -1;
    if (sscanf(p, "%d", &val) != 1) {
	printf("Illegal input.\n");
	return -1;
    }
    if (val < min || val > max) {
	printf("Illegal range.\n");
	return -1;
    }
    return val;
}

void print_list(DISKDEV *dev, PARTBUF *phead)
{
    PARTBUF *pbuf;
    sector_t total, free, used, size;
    int i;
    char name[16];
    char devname[32];

    total = dev->nsect & ~(PS2_PART_MINSIZE - 1);
    free = getfree_partition(dev, phead);
    printf("\nDisk %s: %ldMB total / %ldMB free\n\n",
	   dev->devname, sect2mb(dev, total), sect2mb(dev, free));

    printf("   Device           Size  Partition ID\n");

    used = 0;
    for (i = 1; i <= LINUX_PART_MAX; i++) {
        sprintf(name, LINUX_ID, i);
        if ((pbuf = find_partition(phead, name)) != NULL) {
            size = getsize_partition(dev, pbuf);
            used += size;
            sprintf(devname, "%.16s%d", dev->devname, i);

            printf("%-14s  %6ldMB  %-16.16s",
                   devname, sect2mb(dev, size), pbuf->part->id);
            if (*pbuf->part->name != '\0')
                printf("(%.128s)", pbuf->part->name);
            putchar('\n');
        }
    }

    printf("    ----        %6ldMB  (another system)\n",
           sect2mb(dev, total - free - used));
}

void print_list_expert(DISKDEV *dev, PARTBUF *phead)
{
    PARTBUF *p;
    sector_t total, free;
    int n;

    total = dev->nsect & ~(PS2_PART_MINSIZE - 1);
    free = getfree_partition(dev, phead);
    printf("\nDisk %s: %ldMB total / %ldMB free\n\n",
	   dev->devname, sect2mb(dev, total), sect2mb(dev, free));

    printf("No. Start        Size Num    Total Type Id\n");

    n = 0;
    for (p = phead; p != NULL; p = p->next) {
	printf("%3d %08X %6dMB ", n++,
	       p->part->start, sect2mb(dev, p->part->nsector));
	if (!(p->part->flag & PS2_PART_FLAG_SUB)) {
	    if (p->part->fstype != 0) {
		printf("%3d %6ldMB %4x [%-32.32s]\n",
		       p->part->nsub,
		       sect2mb(dev, getsize_partition(dev, p)),
		       p->part->fstype, p->part->id);
	    } else {
		printf("%3d               <%-32.32s>\n", p->part->nsub, p->part->id);
	    }
	} else {
	    printf("%3d               (%08X)\n", p->part->number, p->part->main);
	}
    }
}

void delete_part(DISKDEV *dev, PARTBUF *phead)
{
    int pno;
    char name[16];
    int result;

    if ((pno = read_num("Partition number", 1, LINUX_PART_MAX)) < 0)
	return;
    sprintf(name, LINUX_ID, pno);
    
    if ((result = remove_partition(dev, phead, name)) < 0)
	print_error(result);
}

void create_part(DISKDEV *dev, PARTBUF *phead)
{
    char *p;
    int pno;
    int fstype;
    char name[17];
    int result;
    int size;

    if ((pno = read_num("Partition number", 1, LINUX_PART_MAX)) < 0)
	return;
    sprintf(name, LINUX_ID, pno);

    printf("Partition fstype([l] linux / [s] linux swap / [v] LVM / [r] reiserfs): ");
    if ((p = read_line()) == NULL)
	return;
    switch (*p) {
    case 'l':
        fstype = LINUX_TYPE;
	break;
    case 's':
        fstype = LINUX_SWAP_TYPE;
	break;
    case 'r':
        fstype = LINUX_REISERFS_TYPE;
        break;
    case 'v':
        fstype = LINUX_LVM_TYPE;
	break;
    default:
	printf("Illegal ID.\n");
	return;
    }

    if ((size = read_num("Partition size (MB)",
			 sect2mb(dev, PS2_PART_MINSIZE),
			 sect2mb(dev, getfree_partition(dev, phead)))) < 0)
	return;
    size = mb2sect(dev, size);

    printf("Comment string : ");
    if ((p = read_line()) == NULL)
	return;

    if ((result = create_partition(dev, phead, name, (sector_t)size,
                                   fstype, p)) < 0)
	print_error(result);
}

void do_interactive(char *device)
{
    DISKDEV *dev;
    PARTBUF *phead;
    char *p;
    int result;
    int expert = 0;

    if ((dev = dev_open(device, 0)) == NULL) {
	perror_fdisk(device);
	exit(1);
    }

    if ((result = read_partition(dev, &phead)) < 0) {
	print_error(result);
	if (result == EP_BADMBR || result == EP_BADPART) {
	    printf("NOTICE: Disk %s seems not to have a PlayStation 2 partition table.\n", dev->devname);
	    printf("Create initialized partition table.\n");
	    format_partition(dev, &phead);
	} else {
	    exit(1);
	}
    }

    while (1) {
	if (!expert)
	    printf("\nCommand (m for help): ");
	else
	    printf("\nExpert command (m for help): ");
	if ((p = read_line()) == NULL)
	    break;

	if (!expert)
	    switch (*p) {
	    case 'p':
		print_list(dev, phead);
		break;
	    case 'd':
		delete_part(dev, phead);
		break;
	    case 'n':
		create_part(dev, phead);
		break;
	    case 'o':
		free_partition(phead);
		format_partition(dev, &phead);
		printf("WARNING: All data in the disk will be lost!!!\n");
		break;
	    case 'w':
		printf("Writing partition table...\n");
		if ((result = write_partition(dev, phead)) < 0) {
		    print_error(result);
		} else {
		    printf("Rereading partition table.\n");
		    reread_partition(dev);
		}
		goto out;
	    case 'q':
		goto out;
	    case 'x':
		expert = 1;
		break;
	    default:
		menu();
	    }
	else
	    switch (*p) {
	    case 'p':
		print_list_expert(dev, phead);
		break;


	    case 'w':
		printf("Writing partition table...\n");
		if ((result = write_partition(dev, phead)) < 0) {
		    print_error(result);
		} else {
		    printf("Rereading partition table.\n");
		    reread_partition(dev);
		}
		goto out;
	    case 'q':
		goto out;
	    case 'r':
		expert = 0;
		break;
	    default:
		menu_expert();
	    }
    }
out:
    free_partition(phead);
    dev_close(dev);
}

void print_list_dev(char *device)
{
    DISKDEV *dev;
    PARTBUF *phead;
    int result;

    if ((dev = dev_open(device, 1)) == NULL) {
	perror_fdisk(device);
	return;
    }
    if ((result = read_partition(dev, &phead)) == 0) {
	print_list(dev, phead);
	free_partition(phead);
    } else {
	printf("Disk %s seems not to have a PlayStation 2 partition table.\n",
	       device);
    }
    dev_close(dev);
}

int cl_create_part(char *device, int size, char *type, char *comment)
{
   DISKDEV *dev;
   PARTBUF *phead;
   int pno;
   int fstype;
   char name[17];
   int result;

   if ((dev = dev_open(device, 0)) == NULL)
   {
      perror_fdisk(device);
      exit(1);
   }

   if ((result = read_partition(dev, &phead)) < 0)
   {
      print_error(result);
      if (result == EP_BADMBR || result == EP_BADPART)
      {
         printf("E: Disk %s seems not to have a PlayStation 2 "
                "partition table.\n", dev->devname);
      }
      exit(1);
   }

   if((pno = next_avail_partition(phead)) >= LINUX_PART_MAX)
   {
      printf("E: No space left in partition table. "
             "No more partitions allowed.\n");
      return 1;
   }
   sprintf(name, LINUX_ID, pno);

   switch (*type)
   {
      case 'l':
         fstype = LINUX_TYPE;
         break;
      case 's':
         fstype = LINUX_SWAP_TYPE;
         break;
      case 'r':
         fstype = LINUX_REISERFS_TYPE;
         break;
      case 'v':
         fstype = LINUX_LVM_TYPE;
         break;
      default:
         printf("E: Illegal filesystem ID.\n");
         return 1;
   }

   size = mb2sect(dev, size);
   if ((result = create_partition(dev, phead, name, (sector_t)size,
                                  fstype, comment)) < 0)
   {
      printf("E: Could not create partition.\n");
      print_error(result);
   }
   else
   {
      if ((result = write_partition(dev, phead)) < 0)
      {
         printf("E: Could not write new partition information to disk.\n");
         print_error(result);
      }
      else
      {
         reread_partition(dev);
         printf("%s%i is the new partition. "
                "Creation successful.\n",device,pno);
      }
   }
   return result;
}

int cl_create_table(char *device)
{
   DISKDEV *dev;
   PARTBUF *phead;
   int result=0;

   if ((dev = dev_open(device, 0)) == NULL)
   {
      perror_fdisk(device);
      exit(1);
   }

   if ((result = read_partition(dev, &phead)) < 0)
   {
      format_partition(dev, &phead);
   }

   free_partition(phead);
   format_partition(dev, &phead);
   if ((result = write_partition(dev, phead)) < 0)
   {
      printf("Partition table creation failed.\n");
      print_error(result);
   }
   else
   {
      reread_partition(dev);
      printf("Partition table creation successful.\n");
   }
   return result;
}

int cl_check_table(char *device)
{
   DISKDEV *dev;
   PARTBUF *phead;
   int result;

   if ((dev = dev_open(device, 0)) == NULL)
   {
      perror_fdisk(device);
      exit(1);
   }

   if (result = read_partition(dev, &phead) < 0)
   {
      printf("No valid PlayStation 2 partition table found.\n");
      return 1;
   }
   else
   {
      printf("Valid PlayStation 2 partition table found.\n");
      return 0;
   }
}

struct option long_options[] = {
   { "check",     no_argument, NULL, 'c' },
   { "add",       no_argument, NULL, 'n' },
   { "create",    no_argument, NULL, 'o' },
   { "help",		no_argument, NULL, 'h' },
   { "list",		no_argument, NULL, 'l' },
   { "size",		no_argument, NULL, 's' },
};

int main(int argc, char **argv)
{
   int do_list = 0, do_size = 0, do_add = 0, do_create = 0, do_check = 0;
   int c;
   int long_index;
   int opts;
   char *device = DEFAULT_DEVICE;

   while ((c = getopt_long(argc, argv, "chlnos", long_options, &long_index)) != EOF)
   {
      switch (c) {
         case 'c':
            do_check = 1;
            break;
         case 'l':
            do_list = 1;
            break;
         case 'n':
            do_add = 1;
            break;   
         case 'o':
            do_create = 1;
            break;
         case 's':
            do_size = 1;
            break;
         case 'h':
         default:
            usage();
      }
   }

   if (do_add)
   {
      opts = argc - optind;
      if (opts != 4)
         usage();

      exit(cl_create_part(argv[optind], atoi(argv[optind+1]),argv[optind+2], argv[optind+3]));
   }

   if (do_check)
   {
      opts = argc - optind;
      if (opts != 1)
         usage();

      exit(cl_check_table(argv[optind]));
   }

   if (do_create)
   {
      opts = argc - optind;
      if (opts != 1)
         usage();

      exit(cl_create_table(argv[optind]));
   }
   
   if (do_list)
   {
      if (optind >= argc)
      {
         print_list_dev(DEFAULT_DEVICE);
         exit(0);
      }
      while (optind < argc) {
         print_list_dev(argv[optind]);
         optind++;
      }
      exit(0);
   }

   if (do_size)
   {
      int fd;
      long size;

      opts = argc - optind;
      if (opts <= 0)
         usage();
      while (optind < argc) {
         if ((fd = open(argv[optind], O_RDONLY)) < 0 ||
             ioctl(fd, BLKGETSIZE, &size) < 0)
         {
            perror_fdisk(argv[optind]);
            exit(1);
         }
         close(fd);
         if (opts == 1)
            printf("%ld\n", size / 2);
         else
            printf("%s: %ld\n", argv[optind], size / 2);
         optind++;
      }
      exit (0);
   }

   if (optind == argc)
      printf("Using %s as default device!\n\n", device);
   else if (optind + 1 == argc)
      device = argv[optind];
   else if (optind + 1 < argc)
      usage();

   do_interactive(device);

   exit (0);
}
